from __future__ import print_function

import os
import sys
from contextlib import contextmanager
from errno import EEXIST, ENOENT, EPERM
from stat import S_IFBLK, S_IFCHR, S_IFIFO, S_IFSOCK, S_IWUSR

from .file import FIFO, BlockDevice, CharacterDevice, RegularFile, Socket, Symlink
from .macro import LOOKUP_INDEX, LOOKUP_OFFSET
from .xattr import has_xattrs, write_xattr

try:
    from os import geteuid
except ImportError:
    _root = False
else:
    _root = geteuid() == 0

try:
    from os import supports_follow_symlinks
except ImportError:
    supports_follow_symlinks = set()

try:
    from os import chown
except ImportError:
    def chown(path, uid, gid):
        pass

try:
    from os import lchown
except ImportError:
    def lchown(path, uid, gid):
        pass

try:
    from os import makedev
except ImportError:
    def makedev(major, minor):
        pass

try:
    from os import mknod
except ImportError:
    def mknod(path, mode=0o600, device=0):
        pass


def set_attributes(pathname, file, set_mode):
    # unsquashfs.c
    os.utime(pathname, (file.time, file.time))
    mode = file.mode
    if _root:
        chown(pathname, file.uid, file.gid)
    else:
        mode &= ~0o6000
    write_xattr(pathname, file)
    if set_mode or (mode & 0o7000):
        try:
            os.chmod(pathname, mode)
        except (IOError, OSError) as e:
            if _root or e.errno != EPERM or not (mode & 0o1000):
                raise
            else:
                os.chmod(pathname, mode & ~0o1000)


def _write_file(file, pathname):
    mode = file.mode
    if not _root and not (mode & S_IWUSR) and has_xattrs(file):
        mode |= S_IWUSR
    flags = os.O_CREAT | os.O_WRONLY
    if sys.platform == "win32":
        flags |= os.O_BINARY
    fd = os.open(pathname, flags, mode & 0o777)  # Use os.open() because umask
    try:
        for block in file.iter_bytes():
            os.write(fd, block)
    finally:
        os.close(fd)


@contextmanager
def appropriate_umask():
    if not _root:
        yield
        return
    old = os.umask(0)
    try:
        yield
    finally:
        os.umask(old)


def unlink(path):
    try:
        os.unlink(path)
    except Exception:
        pass


def write_file(file, pathname, force=False):
    # unsquashfs.c -> write_file
    try:
        os.lstat(pathname)
    except (IOError, OSError) as e:  # Should be FileNotFoundError (Python 2 compatibility)
        if e.errno != ENOENT:
            raise
    else:
        if force:
            os.unlink(pathname)  # Don't ignore errors here.
        elif sys.version_info >= (3, 3):
            raise FileExistsError("file already exists")
        else:
            raise OSError("file already exists")
    with appropriate_umask():
        _write_file(file, pathname)
    set_ = not _root and not (file.mode & S_IWUSR) and has_xattrs(file)
    set_attributes(pathname, file, force or set_)


def _lookup(lookup_table, number):
    # unsquash-34.c
    index = LOOKUP_INDEX(number - 1)
    offset = LOOKUP_OFFSET(number - 1)
    if lookup_table.get(index) is None:
        return None
    return lookup_table[index].get(offset)


def _insert_lookup(lookup_table, number, pathname):
    # unsquash-34.c
    index = LOOKUP_INDEX(number - 1)
    offset = LOOKUP_OFFSET(number - 1)
    if lookup_table.get(index) is None:
        lookup_table[index] = {}
    lookup_table[index][offset] = pathname


def _print(*args, **kwargs):
    if not kwargs.pop("quiet", False):
        print(*args, **kwargs)


def extract_file(file, dest=None, force=False, lookup_table=None, quiet=True):
    # unsquashfs.c -> create_inode
    dest = dest if dest else os.path.basename(file.path)
    _print("extract {} to {}".format(file.path, dest), quiet=quiet)
    lookup_table = lookup_table if lookup_table is not None else {}
    link_path = _lookup(lookup_table, file.inode.inode_number)
    if link_path is not None:
        if force:
            unlink(dest)
        if sys.version_info >= (3, 3) and os.link in supports_follow_symlinks:
            os.link(link_path, dest, follow_symlinks=False)
        return
    if isinstance(file, RegularFile):
        write_file(file, dest, force)
    elif isinstance(file, Symlink):
        if force:
            unlink(dest)
        try:
            os.symlink(file.readlink(), dest)
        except (OSError, AttributeError) as e:
            # Windows + Python < 3.2 -> AttributeError
            # Windows + unprivileged user + Developer Mode disabled -> OSError
            if isinstance(e, OSError) and not (hasattr(e, "winerror") and e.winerror == 1314):
                raise
            elif isinstance(e, AttributeError) and os.path.isfile(dest):
                raise OSError("file already exists")
            with open(dest, "wb") as f:
                f.write(file.inode._symlink)
            os.utime(dest, (file.time, file.time))
        if sys.version_info >= (3, 3) and os.utime in supports_follow_symlinks:
            os.utime(dest, (file.time, file.time), follow_symlinks=False)
        if _root:
            lchown(dest, file.uid, file.gid)
        write_xattr(dest, file)
    elif isinstance(file, (BlockDevice, CharacterDevice)):
        if _root:
            chrdev = isinstance(file, CharacterDevice)
            if force:
                unlink(dest)
            mknod(dest, S_IFCHR if chrdev else S_IFBLK, makedev(file.major, file.minor))
            set_attributes(dest, file, True)
        else:
            _print("WARNING: could not create block or character device because you are not root", quiet=quiet)
    elif isinstance(file, FIFO):
        if force:
            unlink(dest)
        mknod(dest, S_IFIFO, 0)
        set_attributes(dest, file, True)
    elif isinstance(file, Socket):
        mknod(dest, S_IFSOCK, 0)
        set_attributes(dest, file, True)
    else:
        raise Exception("unknown file type")
    _insert_lookup(lookup_table, file.inode.inode_number, dest)


def extract_dir(directory, dest="squashfs-root", force=False, lookup_table=None, quiet=True):
    _print("extract {} to {}".format(directory.path, dest), quiet=quiet)
    lookup_table = lookup_table if lookup_table is not None else {}
    try:
        os.mkdir(dest, 0o700)
    except (IOError, OSError) as e:  # Should be FileExistsError (Python 2 compatibility)
        if force and e.errno == EEXIST:
            os.chmod(dest, 0o700)
        else:
            raise
    for file in directory:
        path = os.path.join(dest, os.path.relpath(file.path, directory.path))
        if file.is_dir:
            extract_dir(file, path, force, lookup_table, quiet)
        else:
            extract_file(file, path, force, lookup_table, quiet)
    set_attributes(dest, directory, True)



from .const import SQUASHFS_CHECK, SQUASHFS_COMPRESSED_BIT, SQUASHFS_COMPRESSED_BIT_BLOCK, SQUASHFS_METADATA_SIZE


def SQUASHFS_COMPRESSED_SIZE(B):
    if B & ~SQUASHFS_COMPRESSED_BIT:
        return B & ~SQUASHFS_COMPRESSED_BIT
    else:
        return SQUASHFS_COMPRESSED_BIT


def SQUASHFS_BIT(flag, bit):
    return ((flag >> bit) & 1) != 0


def SQUASHFS_CHECK_DATA(flags):
    return SQUASHFS_BIT(flags, SQUASHFS_CHECK)


def SQUASHFS_COMPRESSED(B):
    return (B & SQUASHFS_COMPRESSED_BIT) == 0


def SQUASHFS_COMPRESSED_SIZE_BLOCK(B):
    return B & ~SQUASHFS_COMPRESSED_BIT_BLOCK


def SQUASHFS_COMPRESSED_BLOCK(B):
    return (B & SQUASHFS_COMPRESSED_BIT_BLOCK) == 0


def SQUASHFS_INODE_BLK(a):
    return (a >> 16) & 0xFFFFFFFF


def SQUASHFS_INODE_OFFSET(a):
    return a & 0xFFFF


def SQUASHFS_MKINODE(A, B):
    return ((A << 16) + B) & 0xFFFFFFFFFFFFFFFF


def SQUASHFS_MK_VFS_INODE(a, b):
    return ((a << 8) + (b >> 2) + 1) & 0xFFFFFFFF


def SQUASHFS_MODE(a):
    return a & 0xFFF


def SQUASHFS_FRAGMENT_BYTES(A):
    return A * 16


def SQUASHFS_FRAGMENT_INDEX(A):
    return SQUASHFS_FRAGMENT_BYTES(A) // SQUASHFS_METADATA_SIZE


def SQUASHFS_FRAGMENT_INDEX_OFFSET(A):
    return SQUASHFS_FRAGMENT_BYTES(A) % SQUASHFS_METADATA_SIZE


def SQUASHFS_FRAGMENT_INDEXES(A):
    return (SQUASHFS_FRAGMENT_BYTES(A) + SQUASHFS_METADATA_SIZE - 1) // SQUASHFS_METADATA_SIZE


def SQUASHFS_FRAGMENT_INDEX_BYTES(A):
    return SQUASHFS_FRAGMENT_INDEXES(A) * 8


def SQUASHFS_LOOKUP_BYTES(A):
    return A * 8


def SQUASHFS_LOOKUP_BLOCK(A):
    return SQUASHFS_LOOKUP_BYTES(A) // SQUASHFS_METADATA_SIZE


def SQUASHFS_LOOKUP_BLOCK_OFFSET(A):
    return SQUASHFS_LOOKUP_BYTES(A) % SQUASHFS_METADATA_SIZE


def SQUASHFS_LOOKUP_BLOCKS(A):
    return (SQUASHFS_LOOKUP_BYTES(A) + SQUASHFS_METADATA_SIZE - 1) // SQUASHFS_METADATA_SIZE


def SQUASHFS_LOOKUP_BLOCK_BYTES(A):
    return SQUASHFS_LOOKUP_BLOCKS(A) * 8


def SQUASHFS_ID_BYTES(A):
    return A * 4


def SQUASHFS_ID_BLOCK(A):
    return SQUASHFS_ID_BYTES(A) // SQUASHFS_METADATA_SIZE


def SQUASHFS_ID_BLOCK_OFFSET(A):
    return SQUASHFS_ID_BYTES(A) % SQUASHFS_METADATA_SIZE


def SQUASHFS_ID_BLOCKS(A):
    return (SQUASHFS_ID_BYTES(A) + SQUASHFS_METADATA_SIZE - 1) // SQUASHFS_METADATA_SIZE


def SQUASHFS_ID_BLOCK_BYTES(A):
    return SQUASHFS_ID_BLOCKS(A) * 8


def SQUASHFS_XATTR_BYTES(A):
    return A * 16


def SQUASHFS_XATTR_BLOCK(A):
    return SQUASHFS_XATTR_BYTES(A) // SQUASHFS_METADATA_SIZE


def SQUASHFS_XATTR_BLOCK_OFFSET(A):
    return SQUASHFS_XATTR_BYTES(A) % SQUASHFS_METADATA_SIZE


def SQUASHFS_XATTR_BLOCKS(A):
    return (SQUASHFS_XATTR_BYTES(A) + SQUASHFS_METADATA_SIZE - 1) // SQUASHFS_METADATA_SIZE


def SQUASHFS_XATTR_BLOCK_BYTES(A):
    return SQUASHFS_XATTR_BLOCKS(A) * 8


def SQUASHFS_XATTR_BLK(A):
    return (A >> 16) & 0xFFFFFFFF


def SQUASHFS_XATTR_OFFSET(A):
    return A & 0xFFFF


def LOOKUP_INDEX(NUMBER):
    return NUMBER >> 12


def LOOKUP_OFFSET(NUMBER):
    return NUMBER & 0xFFF



import stat
import sys
from ctypes import c_uint16, c_uint32, c_uint64

from ..const import SQUASHFS_INVALID_XATTR
from ..structure import _Base


class InodeHeader(_Base):
    _fields_ = [
        ("_inode_type", c_uint16),
        ("_mode", c_uint16),
        ("_uid", c_uint16),
        ("_guid", c_uint16),
        ("_mtime", c_uint32),
        ("_inode_number", c_uint32),
    ]

    @property
    def inode_type(self):
        return self._inode_type

    @property
    def mode(self):
        return self._mode

    @property
    def uid(self):
        """Index of the UID in the table."""
        return self._uid

    @property
    def guid(self):
        """Index of the GUID in the table."""
        return self._guid

    @property
    def mtime(self):
        return self._mtime

    @property
    def inode_number(self):
        return self._inode_number


class _BaseInode(_Base):

    _header = None
    _uid = None
    _gid = None
    _mode = None

    @property
    def header(self):
        return self._header

    @property
    def uid(self):
        return self._uid

    @property
    def gid(self):
        return self._gid

    @property
    def mode(self):
        return self._mode

    @property
    def type(self):
        return self._header.inode_type

    @property
    def time(self):
        return self._header.mtime

    @property
    def inode_number(self):
        return self._header.inode_number

    @property
    def data(self):
        return 0

    @property
    def xattr(self):
        return getattr(self, "_xattr", SQUASHFS_INVALID_XATTR)

    @property
    def filemode(self):
        if sys.version_info >= (3, 3):
            return stat.filemode(self._mode)
        ret = ['-'] * 10

        if self.is_socket:
            ret[0] = 's'
        if self.is_symlink:
            ret[0] = 'l'
        if self.is_block_device:
            ret[0] = 'b'
        if self.is_dir:
            ret[0] = 'd'
        if self.is_char_device:
            ret[0] = 'c'
        if self.is_fifo:
            ret[0] = 'p'

        if (self._mode & stat.S_IRUSR) == stat.S_IRUSR:
            ret[1] = 'r'
        if (self._mode & stat.S_IWUSR) == stat.S_IWUSR:
            ret[2] = 'w'
        if (self._mode & stat.S_IRGRP) == stat.S_IRGRP:
            ret[4] = 'r'
        if (self._mode & stat.S_IWGRP) == stat.S_IWGRP:
            ret[5] = 'w'
        if (self._mode & stat.S_IROTH) == stat.S_IROTH:
            ret[7] = 'r'
        if (self._mode & stat.S_IWOTH) == stat.S_IWOTH:
            ret[8] = 'w'

        if (self._mode & stat.S_IXUSR) == stat.S_IXUSR:
            ret[3] = 'x'
            if (self._mode & stat.S_ISUID) == stat.S_ISUID:
                ret[3] = 's'
        if (self._mode & stat.S_IXGRP) == stat.S_IXGRP:
            ret[6] = 'x'
            if (self._mode & stat.S_ISGID) == stat.S_ISGID:
                ret[6] = 's'
        if (self._mode & stat.S_IXOTH) == stat.S_IXOTH:
            ret[9] = 'x'

        return ''.join(ret)

    @property
    def is_dir(self):
        return stat.S_ISDIR(self._mode)

    @property
    def is_file(self):
        return stat.S_ISREG(self._mode)

    @property
    def is_symlink(self):
        return stat.S_ISLNK(self._mode)

    @property
    def is_block_device(self):
        return stat.S_ISBLK(self._mode)

    @property
    def is_char_device(self):
        return stat.S_ISCHR(self._mode)

    @property
    def is_fifo(self):
        return stat.S_ISFIFO(self._mode)

    @property
    def is_socket(self):
        return stat.S_ISSOCK(self._mode)


class _DirectoryInodeCommon:

    @property
    def start_block(self):
        return self._start_block

    @property
    def nlink(self):
        return self._nlink

    @property
    def file_size(self):
        return self._file_size

    @property
    def offset(self):
        return self._offset

    @property
    def parent_inode(self):
        return self._parent_inode

    # These fields are either shortcuts or filled in read_inode().

    @property
    def data(self):
        return self.file_size

    @property
    def start(self):
        return self.start_block


class DirectoryInode(_DirectoryInodeCommon, _BaseInode):
    _fields_ = [
        ("_start_block", c_uint32),
        ("_nlink", c_uint32),
        ("_file_size", c_uint16),
        ("_offset", c_uint16),
        ("_parent_inode", c_uint32),
    ]


class ExtendedDirectoryInode(_DirectoryInodeCommon, _BaseInode):
    _fields_ = [
        ("_nlink", c_uint32),
        ("_file_size", c_uint32),
        ("_start_block", c_uint32),
        ("_parent_inode", c_uint32),
        ("_i_count", c_uint16),
        ("_offset", c_uint16),
        ("_xattr", c_uint32),
    ]
    _index = None  # Not used anywhere for now.

    @property
    def i_count(self):
        return self._i_count

    @property
    def index(self):
        return self._index


class _RegularFileInodeCommon:

    _block_list = None
    _frag_bytes = None
    _blocks = None
    _block_start = None
    _block_offset = None
    _sparse = 0

    @property
    def start_block(self):
        return self._start_block

    @property
    def fragment(self):
        return self._fragment

    @property
    def offset(self):
        return self._offset

    @property
    def file_size(self):
        return self._file_size

    @property
    def block_list(self):
        return self._block_list

    # These fields are either shortcuts or filled in read_inode().

    @property
    def data(self):
        return self.file_size

    @property
    def frag_bytes(self):
        return self._frag_bytes

    @property
    def blocks(self):
        return self._blocks

    @property
    def start(self):
        return self.start_block

    @property
    def block_start(self):
        return self._block_start

    @property
    def block_offset(self):
        return self._block_offset

    @property
    def sparse(self):
        return self._sparse


class RegularFileInode(_RegularFileInodeCommon, _BaseInode):
    _fields_ = [
        ("_start_block", c_uint32),
        ("_fragment", c_uint32),
        ("_offset", c_uint32),
        ("_file_size", c_uint32),
    ]


class ExtendedRegularFileInode(_RegularFileInodeCommon, _BaseInode):
    _fields_ = [
        ("_start_block", c_uint64),
        ("_file_size", c_uint64),
        ("_sparse", c_uint64),
        ("_nlink", c_uint32),
        ("_fragment", c_uint32),
        ("_offset", c_uint32),
        ("_xattr", c_uint32),
    ]

    @property
    def sparse(self):
        return self._sparse != 0

    @property
    def nlink(self):
        return self._nlink


class SymlinkInode(_BaseInode):
    _fields_ = [
        ("_nlink", c_uint32),
        ("_symlink_size", c_uint32),
    ]
    _symlink = None

    @property
    def nlink(self):
        return self._nlink

    @property
    def symlink_size(self):
        return self._symlink_size

    @property
    def symlink(self):
        return self._symlink.decode()

    # These fields are either shortcuts or filled in read_inode().

    @property
    def data(self):
        return self.symlink_size


class ExtendedSymlinkInode(SymlinkInode):
    pass


class _DeviceInode(_BaseInode):
    _fields_ = [
        ("_nlink", c_uint32),
        ("_rdev", c_uint32),
    ]

    @property
    def nlink(self):
        return self._nlink

    @property
    def rdev(self):
        return self._rdev

    # These fields are either shortcuts or filled in read_inode().

    @property
    def data(self):
        return self.rdev


class _ExtendedDeviceInode(_DeviceInode):
    _fields_ = [
        ("_xattr", c_uint32),
    ]


class BlockDeviceInode(_DeviceInode):
    pass


class ExtendedBlockDeviceInode(_ExtendedDeviceInode):
    pass


class CharacterDeviceInode(_DeviceInode):
    pass


class ExtendedCharacterDeviceInode(_ExtendedDeviceInode):
    pass


class _IPCInode(_BaseInode):
    _fields_ = [
        ("_nlink", c_uint32),
    ]

    @property
    def nlink(self):
        return self._nlink


class _ExtendedIPCInode(_IPCInode):
    _fields_ = [
        ("_xattr", c_uint32),
    ]


class FIFOInode(_IPCInode):
    pass


class ExtendedFIFOInode(_ExtendedIPCInode):
    pass


class SocketInode(_IPCInode):
    pass


class ExtendedSocketInode(_ExtendedIPCInode):
    pass


inomap = {
    1: DirectoryInode,
    2: RegularFileInode,
    3: SymlinkInode,
    4: BlockDeviceInode,
    5: CharacterDeviceInode,
    6: FIFOInode,
    7: SocketInode,
    8: ExtendedDirectoryInode,
    9: ExtendedRegularFileInode,
    10: ExtendedSymlinkInode,
    11: ExtendedBlockDeviceInode,
    12: ExtendedCharacterDeviceInode,
    13: ExtendedFIFOInode,
    14: ExtendedSocketInode,
}



#!/usr/bin/env python

import argparse
import os
import posixpath
import sys
from datetime import datetime, timedelta
try:
    from datetime import timezone
except ImportError:
    pass
from time import localtime

_is36 = sys.version_info >= (3, 6)
if not _is36:
    from dateutil.tz import tzlocal, tzutc
    UTC = tzutc()
else:
    UTC = timezone.utc

from . import SquashFsImage, __version__
from .const import Compression
from .extract import extract_dir, extract_file
from .file import BlockDevice, CharacterDevice
from .util import find_superblocks

DTFMT = "%Y-%m-%d %H:%M:%S"
ROOT = posixpath.sep


def extract(args):
    with SquashFsImage.from_file(args.file, args.offset) as image:
        file = image.select(args.path)
        if file is None:
            raise Exception("{} not found".format(args.path))
        parent = args.dest or ''
        basename = os.path.basename(file.path) or "squashfs-root"
        dest = os.path.join(parent, basename)
        dirname = os.path.dirname(dest)
        if dirname and not os.path.isdir(dirname):
            os.makedirs(dirname)
        if file.is_dir:
            extract_dir(file, dest, args.force, quiet=args.quiet)
        else:
            extract_file(file, dest, args.force, quiet=args.quiet)


def _dtfromts(timestamp, utc=False):
    if utc:
        tz = UTC
    elif _is36:
        time = localtime(timestamp)
        tz = timezone(timedelta(seconds=time.tm_gmtoff), time.tm_zone)
    else:
        tz = tzlocal()
    return datetime.fromtimestamp(timestamp, tz)


def print_file(file, utc=False, show_tz=False):
    width = 25 - len(str(file.uid)) - len(str(file.gid))
    if isinstance(file, (BlockDevice, CharacterDevice)):
        width = max(0, width - 7)
        data = "{:{width}}{:3d},{:3d}".format(' ', file.major, file.minor, width=width)
    else:
        width = max(0, width)
        data = "{:{width}}".format(file.inode.data, width=width)
    dt = _dtfromts(file.inode.time, utc)
    print("{} {}/{} {} {} {}".format(
        file.filemode,
        file.uid,
        file.gid,
        data,
        dt if show_tz else dt.strftime(DTFMT),
        file.path
    ) + (" -> {}".format(file.readlink()) if file.is_symlink else ''))


def list_(args):
    types = args.type
    if types is not None:
        types = set(''.join(types).replace('f', '-'))
    count = 0
    with SquashFsImage.from_file(args.file, args.offset) as image:
        file = image.select(args.path)
        if file is None:
            raise Exception("{} not found".format(args.path))
        if not file.is_dir:
            print_file(file, args.utc, args.show_tz)
            return
        recursive = not args.recursive if args.path == ROOT else args.recursive
        directory = file.riter() if recursive else file
        if types is None:
            for file in directory:
                print_file(file, args.utc, args.show_tz)
                count += 1
        else:
            for file in directory:
                if file.filemode[0] in types:
                    print_file(file, args.utc, args.show_tz)
                    count += 1
    print("{} file(s) found".format(count))


def scan(args):
    width = 29
    superblocks = find_superblocks(args.file)
    if not superblocks:
        print("No squashfs 4.0 superblock found")
        return
    for idx, superblock in enumerate(superblocks):
        sblk = argparse.Namespace(**superblock)
        dt = _dtfromts(sblk.mkfs_time, args.utc)
        dtstr = dt if args.show_tz else dt.strftime(DTFMT)
        print("Superblock #{}".format(idx + 1))
        print("{:{width}} 0x{:X}".format("Magic:", sblk.s_magic, width=width))
        print("{:{width}} {}".format("Major:", sblk.s_major, width=width))
        print("{:{width}} {}".format("Minor:", sblk.s_minor, width=width))
        print("{:{width}} {}".format("Creation or last append time:", dtstr, width=width))
        print("{:{width}} {}".format("Size:", sblk.bytes_used, width=width))
        print("{:{width}} {}".format("Compression:", Compression(sblk.compression).name, width=width))
        print("{:{width}} {}".format("Block size:", sblk.block_size, width=width))
        print("{:{width}} {}".format("Flags:", sblk.flags, width=width))
        print("{:{width}} {}".format("Number of fragments:", sblk.fragments, width=width))
        print("{:{width}} {}".format("Number of inodes:", sblk.inodes, width=width))
        print("{:{width}} {}".format("Number of ids:", sblk.no_ids, width=width))
        print("{:{width}} 0x{:X}".format("Inode table start:", sblk.inode_table_start, width=width))
        print("{:{width}} 0x{:X}".format("Directory table start:", sblk.directory_table_start, width=width))
        print("{:{width}} 0x{:X}".format("Fragment table start:", sblk.fragment_table_start, width=width))
        print("{:{width}} 0x{:X}".format("Lookup table start:", sblk.lookup_table_start, width=width))
        print("{:{width}} 0x{:X}".format("ID table start:", sblk.id_table_start, width=width))
        print("{:{width}} 0x{:X}".format("xattr ID table start:", sblk.xattr_id_table_start, width=width))
        print("{:{width}} {}".format("Offset:", sblk.offset, width=width))
        if idx != len(superblocks) - 1:
            print()


def main():
    parser = argparse.ArgumentParser(description="Print information about squashfs images.")
    parser.add_argument("-V", "--version", action="version", version="%(prog)s {}".format(__version__))
    subparsers = parser.add_subparsers()  # TODO: required=True Python 3

    pfile = argparse.ArgumentParser(add_help=False)
    pfile.add_argument("file", help="squashfs filesystem")

    poffset = argparse.ArgumentParser(add_help=False)
    poffset.add_argument("-o", "--offset", type=int, default=0, help="absolute position of file system's start. Default: %(default)s")

    ptz = argparse.ArgumentParser(add_help=False)
    ptz.add_argument("--utc", action="store_true", help="use UTC rather than local time zone when displaying time. Default: %(default)s")
    ptz.add_argument("--showtz", action="store_true", dest="show_tz", help="show UTC offset when displaying time. Default: %(default)s")

    helplist = "List the contents of the file system"
    parser_l = subparsers.add_parser("list", parents=[pfile, poffset, ptz], help=helplist.lower(), description=helplist)
    parser_l.add_argument("-p", "--path", default=ROOT, help="absolute path of directory or file to list. Default: %(default)r")
    parser_l.add_argument("-r", "--recursive", action="store_true", help="whether to list recursively. For the root directory the value is inverted. Default: %(default)s")
    parser_l.add_argument("-t", "--type", nargs='+', metavar="TYPE", choices=list("fdlpsbc"), help="when listing a directory, filter by file type with %(choices)s")
    parser_l.set_defaults(func=list_)

    helpextr = "Extract files from the file system"
    parser_e = subparsers.add_parser("extract", parents=[pfile, poffset], help=helpextr.lower(), description=helpextr)
    parser_e.add_argument("-d", "--dest", help="directory that will contain the extracted file(s). If it doesn't exist it will be created. Default: current directory")
    parser_e.add_argument("-p", "--path", default=ROOT, help="absolute path of directory or file to extract. Default: %(default)r")
    parser_e.add_argument("-f", "--force", action="store_true", help="overwrite files that already exist. Default: %(default)s")
    parser_e.add_argument("-q", "--quiet", action="store_true", help="don't print extraction status. Default: %(default)s")
    parser_e.set_defaults(func=extract)

    helpscan = "Find and show all the superblocks that can be found in a file"
    parser_s = subparsers.add_parser("scan", parents=[pfile, ptz], help=helpscan.lower(), description=helpscan)
    parser_s.set_defaults(func=scan)

    args = parser.parse_args()
    if "file" not in args:
        parser.error("the following arguments are required: subcommand")
    if not os.path.isfile(args.file):
        sys.exit("error: file does not exist")
    if "path" in args and not posixpath.isabs(args.path):
        sys.exit("error: path is not absolute")
    if "offset" in args and args.offset is not None and args.offset < 0:
        sys.exit("error: offset cannot be negative")

    try:
        args.func(args)
    except Exception as e:
        sys.exit("error: {}".format(e))


if __name__ == "__main__":
    main()



from .const import Compression


class Compressor:
    name = "none"

    def uncompress(self, src, size, outsize):
        return src


class ZlibCompressor(Compressor):
    name = "gzip"

    def __init__(self):
        import zlib
        self._lib = zlib

    def uncompress(self, src, size, outsize):
        return self._lib.decompress(src)


class LZOCompressor(Compressor):
    name = "lzo"

    def __init__(self):
        import lzo
        self._lib = lzo

    def uncompress(self, src, size, outsize):
        return self._lib.decompress(src, False, outsize)


class LZMACompressor(Compressor):
    name = "lzma"

    def __init__(self):
        try:
            import lzma
        except ImportError:
            from backports import lzma
        self._lib = lzma

    def uncompress(self, src, size, outsize):
        # https://github.com/plougher/squashfs-tools/blob/a04910367d64a5220f623944e15be282647d77ba/squashfs-tools/
        #   lzma_wrapper.c#L40
        # res = LzmaCompress(dest + LZMA_HEADER_SIZE, &outlen, src, size, dest,
        #                    &props_size, 5, block_size, 3, 0, 2, 32, 1);
        # https://github.com/jljusten/LZMA-SDK/blob/781863cdf592da3e97420f50de5dac056ad352a5/C/LzmaLib.h#L96
        # -> level=5, dictSize=block_size, lc=3, lp=0, pb=2, fb=32, numThreads=1
        # https://github.com/plougher/squashfs-tools/blob/a04910367d64a5220f623944e15be282647d77ba/squashfs-tools/
        #   lzma_wrapper.c#L30
        # For some reason, squashfs does not store raw lzma but adds a custom header of 5 B and 8 B little-endian
        # uncompressed size, which can be read with struct.unpack('<Q', src[5:5+8]))
        LZMA_PROPS_SIZE = 5
        LZMA_HEADER_SIZE = LZMA_PROPS_SIZE + 8
        return self._lib.decompress(
            src[LZMA_HEADER_SIZE:],
            format=self._lib.FORMAT_RAW,
            filters=[{"id": self._lib.FILTER_LZMA1, 'lc': 3, 'lp': 0, 'pb': 2}],
        )


class XZCompressor(Compressor):
    name = "xz"

    def __init__(self):
        try:
            import lzma
        except ImportError:
            from backports import lzma
        self._lib = lzma

    def uncompress(self, src, size, outsize):
        return self._lib.decompress(src)


class LZ4Compressor(Compressor):
    name = "lz4"

    def __init__(self):
        import lz4.block
        self._lib = lz4.block

    def uncompress(self, src, size, outsize):
        return self._lib.decompress(src, outsize)


class ZSTDCompressor(Compressor):
    name = "zstd"

    def __init__(self):
        import zstandard
        self._lib = zstandard.ZstdDecompressor()

    def uncompress(self, src, size, outsize):
        return self._lib.decompress(src)


compressors = {
    Compression.NO: Compressor,
    Compression.ZLIB: ZlibCompressor,
    Compression.LZMA: LZMACompressor,
    Compression.LZO: LZOCompressor,
    Compression.XZ: XZCompressor,
    Compression.LZ4: LZ4Compressor,
    Compression.ZSTD: ZSTDCompressor
}



from enum import IntEnum

SQUASHFS_CHECK = 2

SQUASHFS_UIDS = 256
SQUASHFS_GUIDS = 255

SQUASHFS_MAJOR = 4
SQUASHFS_MINOR = 0
SQUASHFS_MAGIC = 0x73717368
SQUASHFS_START = 0

SQUASHFS_METADATA_SIZE = 8192
SQUASHFS_METADATA_LOG = 13

FRAGMENT_BUFFER_DEFAULT = 256
DATA_BUFFER_DEFAULT = 256

SQUASHFS_NAME_LEN = 256
SQUASHFS_INVALID = 0xFFFFFFFFFFFF
SQUASHFS_INVALID_FRAG = 0xFFFFFFFF
SQUASHFS_INVALID_XATTR = 0xFFFFFFFF
SQUASHFS_INVALID_BLK = 0xFFFFFFFFFFFFFFFF  # -1
SQUASHFS_USED_BLK = SQUASHFS_INVALID_BLK - 1  # -2

SQUASHFS_COMPRESSED_BIT = 1 << 15
SQUASHFS_COMPRESSED_BIT_BLOCK = 1 << 24


class Type(IntEnum):
    DIR = 1
    FILE = 2
    SYMLINK = 3
    BLKDEV = 4
    CHRDEV = 5
    FIFO = 6
    SOCKET = 7
    LDIR = 8
    LREG = 9
    LSYMLINK = 10
    LBLKDEV = 11
    LCHRDEV = 12
    LFIFO = 13
    LSOCKET = 14


class Compression(IntEnum):
    NO = 0
    ZLIB = 1
    LZMA = 2
    LZO = 3
    XZ = 4
    LZ4 = 5
    ZSTD = 6



import posixpath
import sys
from collections import OrderedDict

from .const import Type


class File(object):  # Python 2
    """Abstract base class for files."""

    def __init__(self, image, inode, name, parent):
        self._image = image
        self._inode = inode
        self._name = name
        self._parent = parent

    def __str__(self):
        return self.name

    def __repr__(self):
        if sys.version_info >= (3, 3):
            return "{}({!r})".format(self.__class__.__qualname__, self.name)
        else:
            return "{}({!r})".format(self.__class__.__name__, self.name)

    def __lt__(self, other):
        if not isinstance(other, File):
            return NotImplemented
        return self._name < other._name

    def __le__(self, other):
        if not isinstance(other, File):
            return NotImplemented
        return self._name <= other._name

    @property
    def image(self):
        return self._image

    @property
    def inode(self):
        return self._inode

    @property
    def name(self):
        return self._name.decode()

    @property
    def parent(self):
        return self._parent

    @property
    def path(self):
        """Return the file's absolute path."""
        if self._parent is None:
            return '/'
        else:
            return posixpath.join(self._parent.path, self.name)

    @property
    def mode(self):
        return self._inode.mode

    @property
    def uid(self):
        return self._inode.uid

    @property
    def gid(self):
        return self._inode.gid

    @property
    def time(self):
        return self._inode.time

    @property
    def size(self):
        """Return the data attribute of this file's inode.
        - regular file: uncompressed file size in bytes
        - directory: uncompressed listing size in bytes
        - symlink: uncompressed target path size in bytes
        - device file: device number, major and minor
        - FIFO and socket should be 0
        """
        return self._inode.data

    @property
    def xattr(self):
        return self._inode.xattr

    @property
    def filemode(self):
        return self._inode.filemode

    @property
    def is_dir(self):
        return False

    @property
    def is_file(self):
        return False

    @property
    def is_symlink(self):
        return False

    @property
    def is_block_device(self):
        return False

    @property
    def is_char_device(self):
        return False

    @property
    def is_fifo(self):
        return False

    @property
    def is_socket(self):
        return False

    def _lselect(self, lpath, ofs):
        # print lpath,self.name,ofs
        if ofs >= len(lpath):
            return self
        return None


class Directory(File):

    def __init__(self, image, inode, name=b'', parent=None, children=None):
        super(Directory, self).__init__(image, inode, name, parent)  # Python 2
        # OrderedDict for Python 3.6 and lower compatibility.
        self._children = children if children is not None else OrderedDict()

    def __len__(self):
        return len(self._children)

    def __iter__(self):
        return self.iterdir()

    def __getitem__(self, key):
        return self._children[key]

    def __contains__(self, item):
        if isinstance(item, str):
            return item in self._children
        elif isinstance(item, File):
            return item in self._children.values()
        return False

    def __reversed__(self):
        for filename in reversed(self._children):
            yield self._children[filename]

    @property
    def is_dir(self):
        return True

    @property
    def children(self):
        return self._children

    def iterdir(self):
        for file in self._children.values():
            yield file

    def riter(self):
        """Iterate over this directory recursively."""
        yield self
        for file in self._children.values():
            if file.is_dir:
                for f in file.riter():
                    yield f
            else:
                yield file

    def find(self, filename):
        """Find the first file with this name in the subtree."""
        for file in self.riter():
            if file.name == filename:
                return file
        return None

    def select(self, path):
        if path == "/":
            path = ''
        lpath = path.split('/')
        start = self
        ofs = 0
        if not lpath[0]:
            ofs = 1
            while start.parent:
                start = start.parent
        if ofs >= len(lpath):
            return start
        for child_name, child in start.children.items():
            if child_name == lpath[ofs]:
                return child._lselect(lpath, ofs + 1)
        return None

    def _lselect(self, lpath, ofs):
        # print lpath,self.name,ofs
        if ofs >= len(lpath):
            return self
        for child_name, child in self._children.items():
            if child_name == lpath[ofs]:
                return child._lselect(lpath, ofs + 1)
        return None


class RegularFile(File):

    @property
    def is_file(self):
        return True

    def iter_bytes(self):
        return self._image.iter_file(self._inode)

    def read_bytes(self):
        return self._image.read_file(self._inode)

    def read_text(self, encoding="utf8", errors="strict"):
        return self.read_bytes().decode(encoding, errors)


class Symlink(File):

    @property
    def is_symlink(self):
        return True

    def readlink(self):
        return self._inode.symlink


class FIFO(File):

    @property
    def is_fifo(self):
        return True


class Socket(File):

    @property
    def is_socket(self):
        return True


class _Device(File):

    @property
    def major(self):
        return (self._inode.data & 0xFFF00) >> 8

    @property
    def minor(self):
        return (self._inode.data & 0xFF) | ((self._inode.data >> 12) & 0xFFF00)


class CharacterDevice(_Device):

    @property
    def is_char_device(self):
        return True


class BlockDevice(_Device):

    @property
    def is_block_device(self):
        return True


filetype = {
    Type.FILE: RegularFile,
    Type.DIR: Directory,
    Type.SYMLINK: Symlink,
    Type.FIFO: FIFO,
    Type.SOCKET: Socket,
    Type.BLKDEV: BlockDevice,
    Type.CHRDEV: CharacterDevice
}



import sys

from .const import SQUASHFS_INVALID_BLK, SQUASHFS_INVALID_XATTR


def has_xattrs(file):
    # unsquashfs_xattr.c
    return (
        file.xattr != SQUASHFS_INVALID_XATTR
        and file.image.sblk.xattr_id_table_start != SQUASHFS_INVALID_BLK
    )


def write_xattr(pathname, file):
    # unsquashfs_xattr.c
    if sys.version_info < (3, 3) or sys.platform != "linux" or not has_xattrs(file):
        return



# -*- coding: utf-8 -*-

"""
This module is released with the LGPL license.
Copyright 2011-2012

Matteo Mattei <matteo.mattei@gmail.com>
Nicola Ponzeveroni <nicola.ponzeveroni@gilbarco.com>

It is intended to be used to access files inside a Squashfs 4.0 little endian
image file.

Based on squashfs-tools by Phillip Lougher <phillip@squashfs.org.uk>

https://github.com/matteomattei/PySquashfsImage
https://github.com/plougher/squashfs-tools

"""
__all__ = ["SquashFsImage"]
__version__ = "0.9.0"

import io
import stat
import struct
import sys
from ctypes import sizeof
try:
    from functools import lru_cache
except ImportError:
    def lru_cache(maxsize=128, typed=False):
        return lambda user_function: user_function

from .compressor import compressors
from .const import (
    SQUASHFS_INVALID_BLK,
    SQUASHFS_INVALID_FRAG,
    SQUASHFS_METADATA_SIZE,
    Type,
)
from .file import Directory, filetype
from .macro import (
    SQUASHFS_CHECK_DATA,
    SQUASHFS_COMPRESSED,
    SQUASHFS_COMPRESSED_BLOCK,
    SQUASHFS_COMPRESSED_SIZE,
    SQUASHFS_COMPRESSED_SIZE_BLOCK,
    SQUASHFS_FRAGMENT_BYTES,
    SQUASHFS_FRAGMENT_INDEXES,
    SQUASHFS_ID_BLOCK_BYTES,
    SQUASHFS_ID_BLOCKS,
    SQUASHFS_ID_BYTES,
    SQUASHFS_INODE_BLK,
    SQUASHFS_INODE_OFFSET,
    SQUASHFS_XATTR_BLOCK_BYTES,
    SQUASHFS_XATTR_BLOCKS,
    SQUASHFS_XATTR_BYTES,
)
from .structure import DirEntry, DirHeader, FragmentEntry, Superblock, XattrId, XattrTable
from .structure.inode import InodeHeader, inomap
from .util import check_super


SQUASHFS_LOOKUP_TYPE = [
    0,
    stat.S_IFDIR,
    stat.S_IFREG,
    stat.S_IFLNK,
    stat.S_IFBLK,
    stat.S_IFCHR,
    stat.S_IFIFO,
    stat.S_IFSOCK,
    stat.S_IFDIR,
    stat.S_IFREG,
    stat.S_IFLNK,
    stat.S_IFBLK,
    stat.S_IFCHR,
    stat.S_IFIFO,
    stat.S_IFSOCK
]


class SquashFsImage(object):

    def __init__(self, fd, offset=0, closefd=True):
        self._fd = fd
        self._offset = offset
        self._closefd = closefd
        self._sblk = None
        self._root = None
        self._comp = None
        self._inode_table_hash = {}
        self._directory_table_hash = {}
        self._fragment_table = []
        self._id_table = {}
        self._hash_table = {}
        self._xattrs = b""
        self._initialize()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._closefd:
            self.close()

    def __iter__(self):
        return self._root.riter()

    @property
    def root(self):
        return self._root

    @property
    def sblk(self):
        return self._sblk

    @property
    def size(self):
        """Filesystem size in bytes."""
        return self._sblk.bytes_used

    @classmethod
    def from_bytes(cls, bytes_, offset=0):
        return cls(io.BytesIO(bytes_), offset)

    @classmethod
    def from_file(cls, path, offset=0):
        return cls(open(path, "rb"), offset)

    def close(self):
        self._fd.close()
        self._fd = None

    def _read_super(self):
        self._sblk = Superblock.from_fd(self._fd)
        if not check_super(self._sblk):
            raise IOError("The file supplied is not a squashfs 4.0 image")
        self._comp = self._get_compressor(self._sblk.compression)

    def _get_compressor(self, compression_id):
        if compression_id not in compressors:
            raise ValueError("Unknown compression method %r" % compression_id)
        return compressors[compression_id]()

    def _initialize(self):
        self._fd.seek(self._offset)
        self._read_super()
        self._read_uids_guids()
        self._read_fragment_table()
        self._read_xattrs_from_disk()
        root_block = SQUASHFS_INODE_BLK(self._sblk.root_inode)
        root_offs = SQUASHFS_INODE_OFFSET(self._sblk.root_inode)
        self._root = self._dir_scan(root_block, root_offs)

    @lru_cache(maxsize=256)
    def _read_data_block(self, start, size):
        c_byte = SQUASHFS_COMPRESSED_SIZE_BLOCK(size)
        self._fd.seek(self._offset + start)
        data = self._fd.read(c_byte)
        if SQUASHFS_COMPRESSED_BLOCK(size):
            return self._comp.uncompress(data, c_byte, self._sblk.block_size)
        else:
            return data

    def iter_file(self, inode):
        # unsquashfs.c -> write_file
        start = inode.start
        file_end = inode.data // self._sblk.block_size
        if inode.blocks:
            block_list = self._read_block_list(inode.block_start, inode.block_offset, inode.blocks)
            for i, block in enumerate(block_list):
                if block == SQUASHFS_INVALID_FRAG:
                    continue
                if block:  # non sparse file
                    yield self._read_data_block(start, block)
                    start += SQUASHFS_COMPRESSED_SIZE_BLOCK(block)
                else:
                    if i == file_end:
                        yield b'\x00' * (inode.data & (self._sblk.block_size - 1))
                    else:
                        yield b'\x00' * self._sblk.block_size
        if inode.frag_bytes:
            start, size = self._read_fragment(inode.fragment)
            buffer = self._read_data_block(start, size)
            yield buffer[inode.offset : inode.offset + inode.frag_bytes]

    def read_file(self, inode):
        return b''.join(self.iter_file(inode))

    def _read_block_list(self, start, offset, blocks):
        # unsquash-4.c
        size = 4  # sizeof(unsigned int)
        idata, _, _ = self._read_inode_data(start, offset, blocks * size)
        ret = []
        ofs = 0
        for _ in range(blocks):  # use struct.iter_unpack when Py3 only
            ret.append(self._make_buf_integer(idata, ofs, size))
            ofs += size
        return ret

    def _read_block(self, start, expected=SQUASHFS_METADATA_SIZE):
        """Read a block starting at offset `start` relative to the start of the image.

        Return the uncompressed block and the start of the next compressed one.
        """
        # unsquashfs.c
        self._fd.seek(self._offset + start)
        c_byte = self._read_short()
        offset = 3 if SQUASHFS_CHECK_DATA(self._sblk.flags) else 2
        self._fd.seek(self._offset + start + offset)
        size = SQUASHFS_COMPRESSED_SIZE(c_byte)
        block = self._fd.read(size)
        if SQUASHFS_COMPRESSED(c_byte):
            block = self._comp.uncompress(block, size, expected)
        return block, start + offset + size

    def _read_fragment_table(self):
        # unsquash-4.c
        bytes_ = SQUASHFS_FRAGMENT_BYTES(self._sblk.fragments)
        indexes = SQUASHFS_FRAGMENT_INDEXES(self._sblk.fragments)
        if self._sblk.fragments == 0:
            return
        self._fd.seek(self._offset + self._sblk.fragment_table_start)
        fragment_table_index = [self._read_long() for _ in range(indexes)]
        table = b''
        for i, index in enumerate(fragment_table_index):
            if (i + 1) != indexes:
                expected = SQUASHFS_METADATA_SIZE
            else:
                expected = bytes_ & (SQUASHFS_METADATA_SIZE - 1)
            table += self._read_block(index, expected)[0]
        ofs = 0
        while ofs < len(table):
            entry = FragmentEntry.from_bytes(table, ofs)
            ofs += sizeof(FragmentEntry)
            self._fragment_table.append(entry)

    def _read_fragment(self, fragment):
        # unsquash-4.c
        entry = self._fragment_table[fragment]
        return (entry.start_block, entry.size)

    def _read_inode(self, start_block, offset):
        # unsquash-4.c
        start = self._sblk.inode_table_start + start_block
        idata, start, offset = self._read_inode_data(start, offset, sizeof(InodeHeader))
        header = InodeHeader.from_bytes(idata)
        cls = inomap[header.inode_type]
        idata, start, offset = self._read_inode_data(start, offset, sizeof(cls))
        ino = cls.from_bytes(idata)
        ino._header = header
        ino._uid = self._id_table[header.uid]
        ino._gid = self._id_table[header.guid]
        ino._mode = SQUASHFS_LOOKUP_TYPE[header.inode_type] | header.mode
        try:
            inode_type = Type(header.inode_type)
        except ValueError:
            raise RuntimeError("Unknown inode type %d in read_inode!\n" % header.inode_type)
        if inode_type in (Type.FILE, Type.LREG):
            ino._block_start = start
            ino._block_offset = offset
            if ino.fragment == SQUASHFS_INVALID_FRAG:
                ino._frag_bytes = 0
                ino._blocks = (ino.data + self._sblk.block_size - 1) >> self._sblk.block_log
            else:
                ino._frag_bytes = ino.file_size % self._sblk.block_size
                ino._blocks = ino.data >> self._sblk.block_log
        elif inode_type in (Type.SYMLINK, Type.LSYMLINK):
            idata, start, offset = self._read_inode_data(start, offset, ino.symlink_size)
            ino._symlink = idata
            if inode_type == Type.LSYMLINK:
                idata, start, offset = self._read_inode_data(start, offset, 4)
                ino._xattr = self._make_buf_integer(idata, 0, len(idata))
        return ino

    def _opendir(self, block_start, offset):
        # unsquash-4.c -> squashfs_opendir
        inode = self._read_inode(block_start, offset)
        directory = Directory(self, inode)
        directory.entries = []
        if inode.data == 3:
            return directory
        start = self._sblk.directory_table_start + inode.start
        offset = inode.offset
        size = inode.data - 3
        bytes_ = 0
        while bytes_ < size:
            ddata, start, offset = self._read_directory_data(start, offset, sizeof(DirHeader))
            dirh = DirHeader.from_bytes(ddata)
            bytes_ += sizeof(DirHeader)
            for _ in range(dirh.count + 1):
                ddata, start, offset = self._read_directory_data(start, offset, sizeof(DirEntry))
                dire = DirEntry.from_bytes(ddata)
                namelen = dire.size + 1
                ddata, start, offset = self._read_directory_data(start, offset, namelen)
                dire._name = ddata
                bytes_ += sizeof(DirEntry) + namelen
                directory.entries.append({
                    "name": dire._name,
                    "start_block": dirh.start_block,
                    "offset": dire.offset,
                    "type": dire.type
                })
        return directory

    def _read_uids_guids(self):
        size = 4
        bytes_ = SQUASHFS_ID_BYTES(self._sblk.no_ids)
        indexes = SQUASHFS_ID_BLOCKS(self._sblk.no_ids)
        id_index_table = []
        self._fd.seek(self._offset + self._sblk.id_table_start)
        for _ in range(indexes):
            id_index_table.append(self._make_integer(SQUASHFS_ID_BLOCK_BYTES(1)))
        for i, idx in enumerate(id_index_table):
            if (i + 1) != indexes:
                expected = SQUASHFS_METADATA_SIZE
            else:
                expected = bytes_ & (SQUASHFS_METADATA_SIZE - 1)
            block = self._read_block(idx, expected)[0]
            offset = 0
            index = i * (SQUASHFS_METADATA_SIZE // 4)
            while offset < len(block):
                self._id_table[index] = self._make_buf_integer(block, offset, size)
                offset += size
                index += 1

    def _read_xattrs_from_disk(self):
        # read_xattrs.c
        if self._sblk.xattr_id_table_start == SQUASHFS_INVALID_BLK:
            return SQUASHFS_INVALID_BLK
        self._fd.seek(self._offset + self._sblk.xattr_id_table_start)
        id_table = XattrTable.from_fd(self._fd)
        ids = id_table.xattr_ids
        xattr_table_start = id_table.xattr_table_start
        indexes = SQUASHFS_XATTR_BLOCKS(ids)
        index = []
        for _ in range(indexes):
            index.append(self._make_integer(SQUASHFS_XATTR_BLOCK_BYTES(1)))
        bytes_ = SQUASHFS_XATTR_BYTES(ids)
        xattr_ids = {}
        for i, idx in enumerate(index):
            if (i + 1) != indexes:
                expected = SQUASHFS_METADATA_SIZE
            else:
                expected = bytes_ & (SQUASHFS_METADATA_SIZE - 1)
            block = self._read_block(idx, expected)[0]
            cur_idx = (i * SQUASHFS_METADATA_SIZE) / 16
            ofs = 0
            while ofs < len(block):
                xattr_ids[cur_idx] = XattrId.from_bytes(block, ofs)
                cur_idx += 1
                ofs += sizeof(XattrId)
        start = xattr_table_start
        i = 0
        while start < index[0]:
            self._hash_table[start] = i * SQUASHFS_METADATA_SIZE
            block, start = self._read_block(start)
            for i in range(len(block), SQUASHFS_METADATA_SIZE):
                block += b'\x00'
            self._xattrs += block
            i += 1
        return ids

    def _dir_scan(self, start_block, offset):
        directory = self._opendir(start_block, offset)
        for entry in directory.entries:  # No need for squashfs_readdir()
            start_block = entry["start_block"]
            offset = entry["offset"]
            if entry["type"] == Type.DIR:
                subdir = self._dir_scan(start_block, offset)
                subdir._parent = directory
                subdir._name = entry["name"]
                directory.children[subdir.name] = subdir
            else:
                inode = self._read_inode(start_block, offset)
                cls = filetype[entry["type"]]
                file = cls(self, inode, entry["name"], directory)
                directory.children[file.name] = file
        del directory.entries
        return directory

    def _make_integer(self, length):
        """Assemble multibyte integer."""
        return self._make_buf_integer(self._fd.read(length), 0, length)

    def _make_buf_integer(self, buf, start, length):
        """Assemble multibyte integer."""
        if sys.version_info < (3, 2):
            ret = 0
            pwr = 1
            for i in range(start, start + length):
                ret += (ord(buf[i]) & 0xFF) * pwr
                pwr *= 0x100
            return ret
        else:
            return int.from_bytes(buf[start : start + length], byteorder='little')

    def _read_integer(self, fmt):
        return struct.unpack(fmt, self._fd.read(struct.calcsize(fmt)))[0]

    def _read_short(self):
        return self._read_integer("<H")

    def _read_long(self):
        return self._read_integer("<Q")

    def _get_metadata(self, hash_table, start):
        try:
            return hash_table[start]
        except KeyError:
            hash_table[start] = dict(zip(("buffer", "next_index"), self._read_block(start)))
            return hash_table[start]

    def _read_metadata(self, hash_table, block, offset, length):
        data = b''
        while True:
            entry = self._get_metadata(hash_table, block)
            copy = len(entry["buffer"]) - offset
            if copy < length:
                data += entry["buffer"][offset:]
                length -= copy
                block = entry["next_index"]
                offset = 0
            elif copy == length:
                data += entry["buffer"][offset : offset + length]
                return data, entry["next_index"], 0
            else:
                data += entry["buffer"][offset : offset + length]
                return data, block, offset + length

    def _read_inode_data(self, block, offset, length):
        return self._read_metadata(self._inode_table_hash, block, offset, length)

    def _read_directory_data(self, block, offset, length):
        return self._read_metadata(self._directory_table_hash, block, offset, length)

    def find(self, filename):
        return self._root.find(filename)

    def select(self, path):
        return self._root.select(path)



from ctypes import LittleEndianStructure, c_int16, c_uint16, c_uint32, c_uint64, sizeof

from ..const import Type


class _Base(LittleEndianStructure):

    def __iter__(self):
        # This allows calling dict() on instances of this class.
        for name, _ in self._fields_:
            name = name.lstrip('_')
            yield name, getattr(self, name)

    @classmethod
    def from_bytes(cls, bytes_, offset=0):
        return cls.from_buffer_copy(bytes_, offset)
    
    @classmethod
    def from_fd(cls, fd):
        return cls.from_bytes(fd.read(sizeof(cls)))


class Superblock(_Base):
    _fields_ = [
        ("_s_magic", c_uint32),
        ("_inodes", c_uint32),
        ("_mkfs_time", c_uint32),
        ("_block_size", c_uint32),
        ("_fragments", c_uint32),
        ("_compression", c_uint16),
        ("_block_log", c_uint16),
        ("_flags", c_uint16),
        ("_no_ids", c_uint16),
        ("_s_major", c_uint16),
        ("_s_minor", c_uint16),
        ("_root_inode", c_uint64),
        ("_bytes_used", c_uint64),
        ("_id_table_start", c_uint64),
        ("_xattr_id_table_start", c_uint64),
        ("_inode_table_start", c_uint64),
        ("_directory_table_start", c_uint64),
        ("_fragment_table_start", c_uint64),
        ("_lookup_table_start", c_uint64)
    ]

    @property
    def s_magic(self):
        return self._s_magic
    
    @property
    def inodes(self):
        return self._inodes
    
    @property
    def mkfs_time(self):
        return self._mkfs_time
    
    @property
    def block_size(self):
        return self._block_size
    
    @property
    def fragments(self):
        return self._fragments
    
    @property
    def compression(self):
        return self._compression
    
    @property
    def block_log(self):
        return self._block_log
    
    @property
    def flags(self):
        return self._flags
    
    @property
    def no_ids(self):
        return self._no_ids
    
    @property
    def s_major(self):
        return self._s_major
    
    @property
    def s_minor(self):
        return self._s_minor
    
    @property
    def root_inode(self):
        return self._root_inode
    
    @property
    def bytes_used(self):
        return self._bytes_used
    
    @property
    def id_table_start(self):
        return self._id_table_start
    
    @property
    def xattr_id_table_start(self):
        return self._xattr_id_table_start
    
    @property
    def inode_table_start(self):
        return self._inode_table_start
    
    @property
    def directory_table_start(self):
        return self._directory_table_start
    
    @property
    def fragment_table_start(self):
        return self._fragment_table_start
    
    @property
    def lookup_table_start(self):
        return self._lookup_table_start


class DirEntry(_Base):
    _fields_ = [
        ("_offset", c_uint16),
        ("_inode_number", c_int16),
        ("_type", c_uint16),
        ("_size", c_uint16),
    ]
    _name = None

    @property
    def offset(self):
        return self._offset

    @property
    def inode_number(self):
        return self._inode_number

    @property
    def type(self):
        return Type(self._type)

    @property
    def size(self):
        return self._size

    @property
    def name(self):
        return self._name.decode()


class DirHeader(_Base):
    _fields_ = [
        ("_count", c_uint32),
        ("_start_block", c_uint32),
        ("_inode_number", c_uint32),
    ]

    @property
    def count(self):
        return self._count

    @property
    def start_block(self):
        return self._start_block

    @property
    def inode_number(self):
        return self._inode_number


class FragmentEntry(_Base):
    _fields_ = [
        ("_start_block", c_uint64),
        ("_size", c_uint32),
        ("_unused", c_uint32)
    ]
    
    @property
    def start_block(self):
        return self._start_block
    
    @property
    def size(self):
        return self._size
    
    @property
    def unused(self):
        return self._unused


class XattrId(_Base):
    _fields_ = [
        ("_xattr", c_uint64),
        ("_count", c_uint32),
        ("_size", c_uint32)
    ]
    
    @property
    def xattr(self):
        return self._xattr
    
    @property
    def count(self):
        return self._count
    
    @property
    def size(self):
        return self._size


class XattrTable(_Base):
    _fields_ = [
        ("_xattr_table_start", c_uint64),
        ("_xattr_ids", c_uint32),
        ("_unused", c_uint32)
    ]

    @property
    def xattr_table_start(self):
        return self._xattr_table_start
    
    @property
    def xattr_ids(self):
        return self._xattr_ids
    
    @property
    def unused(self):
        return self._unused



import io
from functools import partial

from .const import SQUASHFS_MAGIC, Compression
from .structure import Superblock


try:
    MAGIC_BYTES = SQUASHFS_MAGIC.to_bytes(4, "little")
except AttributeError:
    MAGIC_BYTES = b'hsqs'  # Python 2


def check_super(sblk):
    if sblk.s_magic != SQUASHFS_MAGIC or sblk.s_major != 4 or sblk.s_minor != 0:
        return False
    if not min(Compression) <= sblk.compression <= max(Compression):
        return False
    return True


def _find_superblocks(stream, size=1024**2):
    stream.seek(0)
    indexes = set()
    result = []
    prev_block = b''
    for count, next_block in enumerate(iter(partial(stream.read, size), b'')):
        # We don't want to "cut" in the middle of a magic.
        block = prev_block + next_block
        index = block.find(MAGIC_BYTES)
        if index != -1:
            indexes.add(index + (count * size) - len(prev_block))
        prev_block = next_block[-(len(MAGIC_BYTES) - 1) :]
    for index in sorted(indexes):
        stream.seek(index)
        sblk = Superblock.from_fd(stream)
        if check_super(sblk):
            result.append(dict(sblk, offset=index))
    return result


def find_superblocks(file_or_bytes, size=1024**2):
    """Return a list of dictionaries representing the
    superblocks found in the file with their offset.
    """
    if hasattr(file_or_bytes, "read"):
        return _find_superblocks(file_or_bytes, size)
    try:
        with open(file_or_bytes, "rb") as f:
            return _find_superblocks(f, size)
    except (IOError, OSError, TypeError, UnicodeDecodeError):
        # TypeError and IOError: Python 2 only
        # UnicodeDecodeError: Python 3 only, when argument is file as bytes
        pass
    try:
        return _find_superblocks(io.BytesIO(file_or_bytes), size)
    except TypeError:
        pass
    raise Exception("file not found or wrong type")  # TODO: Python 3 from None
